<iframe src="../.ibox.html?raw=true" style="border:none; position:fixed; width:40px; right:0; z-index=999;"></iframe>

# Syntax

`ogma` predominantly uses a **prefix** notation syntax with a focus on **pipelining** output
into the next input block. It has a type system which falls between dynamic and static, and
supports user defined commands (**definitions**) and types.

## Structure
---
For a given **block** of code, the syntax is `cmd arg1 arg2`; that is,
a **command**, followed by _zero or more_ **arguments** separated by _spaces_.
The _input_ into the block comes from the _output_ of the previous block. Blocks can be
chained together into an **expression** by using the pipe character (`|`).
A simple example shows the different parts:
```plaintext
\ 'Hello' | + ', ' | + 'world!'
^           ^        ^          --> Commands
  ^-----^     ^--^     ^------^ --> Args
^-------^   ^----^   ^--------^ --> Blocks
^-----------------------------^ --> Expression 
```

The syntax might be initially unfamiliar, especially when working with arithmetic
expressions, due to the use of prefix notation. The notation does however allow for less ambiguity. Take
for example the mathematical expression: `3 * 4 - 4 / 2`. The order of operations defines
how this should be evaluated, but this does create parsing and evaluation ambiguity unless
explicitly defined. The order of operations is an _implicit_ assumption based on
collective math education. In ogma the expression would be defined: `\ 3 | * 4 | - {\4 |
/ 2}`. Notice the `{\4 | / 2}` part, this is a **sub-expression**.

## Argument types and Sub-expressions
---
Arguments can be specified as literals (strings, numbers, booleans, nil), variables, or
_expressions_. Strings are encased in matching **quote** or **double quote** characters, and these quotes can
be skipped if the string is just a _single word_ (no spaces). Booleans and Nils use the pound character to flag their
value. Variables lead with a dollar sign symbol. Expressions are defined by wrapping the expression in
curly braces.

| Argument | Examples         |
| -------- | ---------------- |
| Number   | `0`, `-3`, `3.14`, `1e6` |
| String   | `foo`, `'foo bar'`, `"Hello, world!"` |
| Boolean  | `#t`, `#f` (true, false)|
| Nil      | `#n` |
| Input    | `#i` |
| Variable | `$x`, `$foo`, `$foo-bar` |

Sub-expressions are just regular expressions encased in curly braces. They can be nested
arbitrarily deep, and generally take on the _input_ of the _block they are defined in_. For
example, this expression uses sub-expressions to branch on conditionals:

```plaintext
\ 10 | if {mod 3 | = 0} 'divisible by 3'
	  {mod 5 | = 0} 'divisible by 5'
	  {mod 7 | = 0} 'divisible by 7'
	  'something else'
```

Each sub-expression calculates the modulus of the _input_ (which is 10) to a number (3,5,7) and
checks if it equals zero.

## Sub-expressions without the braces
---
There exists a shorthand for sub-expressions which can be used as the **_last_** argument. The
shorthand allows one to elide braces enclosing an expression, instead relying on
the right-associativity of arguments. This means that a block which takes the form `cmd1 arg1 {cmd2
arg2}` where the 2nd argument into `cmd1` is an expression, can be written as `cmd1 arg1 cmd2 arg2`.
This nests, such that `cmd1 arg1 {cmd2 arg2 {cmd3 arg3}} = cmd1 arg1 cmd2 arg2 cmd3 arg3`. This has
some important details:
1. Only the trailing argument can use this syntax,
2. A pipe character still defines the boundary to the shallowest block.

For instance, `cmd1 arg1 { cmd2 arg2 | cmd3 } ≠ cmd1 arg1 cmd2 arg2 | cmd3`.
Instead, `  cmd1 arg1 cmd2 arg2 | cmd3 = cmd1 arg1 { cmd2 arg2 } | cmd3`.

> ℹ️ Try not to overuse the sub-expression shorthand.
> Sometimes encasing a sub-expression with braces will make your code more legible!

## User defined definitions
---
Users can define their own implementations and types! These definitions have specific syntax which
is described in the [User Definitions](../07%20User%20Definitions?book=true) chapter.

