<iframe src="../.ibox.html?raw=true" style="border:none; position:fixed; width:40px; right:0; z-index=999;"></iframe>

# User Defined Implementations

ogma supports encapsulating expressions through **implementations**. These implementations are
_defined_ using the `def` command. A definition is _similar_ to a function in other languages, but
not the same. In ogma, a definition's output can be dependent on types of the input and arguments.

## Syntax
---
The `def` command takes the name of the implementation, an optional
input type to apply to, a list of parameters (optionally typed) and then the expression body.
```plaintext
Usage:
  => def name [Ty] ([param1[:Ty|Expr]] [param2[:Ty|Expr]] ...) { expr }
```

Parameters can be typed `:Expr`, which tells ogma that the parameter should **not** be evaluated
before being passed into the definition. This parameter type would be used for _closure_
expressions that are passed to commands such as `map` or `fold`.

The most basic definition could be defined as `def return-5 () { \5 }`. This simply returns the
number 5. Say one wanted to add 5 to a parameter? `def add-5 (lhs) { \$lhs | + 5 }`. This is bad
style, instead one should leverage the _data pipeline_ to define the definitions: `def add-5 () { +
5 }`. This example should also be typed: `def add-5 Num () { + 5 }`. Typing the input means calling
`add-5` on an input that is **not** a number would result in a semantics error.

> It is recommended to always type implementation inputs and parameters. This provides better error
> checking and readability.

## Overloading on input type
---
Implementations can be _overloaded_. This is usually done for user-defined types (implementing
`eq` or `cmp`), but can also be done on _any_ command (without interfering with `ogma` commands).
For instance, the previous `add-5` example only worked on `Num` inputs. Say one wanted to extend it
to work with `Str`ing inputs, adding a stringified 5 to the string: `def add-5 Str () { + {\5|to-str} }`.
Now `add-5` has two _implementations_ differentiated by the input type.

These implementations show up in the definition list:

![](../assets/usr-def.impls.gif?raw=true)

> Implementations can only be overloaded on the input type, not the arity of the parameters for the
> _same_ input type.


Many supplied commands in `ogma` are defined using this style. For example, the `last` command is
defined twice whether it has a string input or table input. The table input takes an extra
parameter:

![](../assets/usr-def.impls2.png?raw=true)

When evaluating an implementation, _the parameters are eagerly evaluated_, meaning the parameter
can be used like a _value_. There are cases where this is not preferred, instead an _expression_ is
to be passed through and evaluated in a different context. Usually this occurs when working with
table commands that take an expression that gets evaluated with a `TableRow` input. Parameters are
assumed to be eagerly evaluated unless explicitly defined with the `Expr` type.

> When using the `Expr` type, a parameter will be **evaluated outside of the context it was defined
> in**. The exact context depends on the semantics of the expression using the parameter.
> Importantly, the _input_ into the parameter can turn out to be **not** the input that was
> available where the expression was defined. This can lead to subtle type errors.
>
> It is recommended to only use `Expr` types when necessary and prefer eager types as priority.

